from itertools import cycle

import moka_py


def test_bench_set(benchmark):
    moka = moka_py.Moka(100_000)
    to_set = cycle(iter(list(range(100_000))))

    def _set():
        k = next(to_set)
        moka.set(k, k)

    benchmark(_set)


def test_bench_set_huge(benchmark):
    moka = moka_py.Moka(10_000)
    to_set = cycle(iter(list(range(10_000))))
    payload = "hello" * 100_000

    def _set():
        k = next(to_set)
        moka.set(k, payload)

    benchmark(_set)


def test_bench_get(benchmark):
    moka = moka_py.Moka(10_000)
    payload = "hello" * 100_000
    for key in range(10_000):
        moka.set(f"pretty_long_key_of_index_{key}", payload)

    def _bench():
        moka.get("pretty_long_key_of_index_5432")

    benchmark(_bench)


def test_bench_get_non_existent(benchmark):
    moka = moka_py.Moka(10_000)
    payload = "hello" * 100_000
    for key in range(10_000):
        moka.set(f"pretty_long_key_of_index_{key}", payload)

    def _bench():
        moka.get("hello")

    benchmark(_bench)


def test_bench_get_with(benchmark):
    moka = moka_py.Moka(10_000)

    def init():
        return 5

    def _bench():
        moka.get_with("hello", init)

    benchmark(_bench)



import asyncio
from threading import Thread
from time import sleep

import pytest

import moka_py
import random


def test_decorator_sync():
    lst = []

    @moka_py.cached()
    def f(x: int, y: int) -> float:
        lst.append((x, y))
        return x / y

    cases = [
        (1, 2), (3, 4), (5, 6), (7, 8)
    ]

    answers = {}

    for _ in range(100):
        copy = cases[:]
        random.shuffle(copy)
        for case in copy:
            result = f(*case)
            if case not in answers:
                answers[case] = result
            else:
                assert answers[case] == result

    # A call with the same arguments occurs only once
    assert len(lst) == len(cases)


@pytest.mark.asyncio
async def test_decorator_async():
    lst = []

    @moka_py.cached()
    async def f(x: int, y: int) -> float:
        lst.append((x, y))
        await asyncio.sleep(0.01)
        return x / y

    cases = [
        (1, 2), (3, 4), (5, 6), (7, 8)
    ]

    answers = {}

    for _ in range(100):
        copy = cases[:]
        random.shuffle(copy)
        for case in copy:
            result = await f(*case)
            if case not in answers:
                answers[case] = result
            else:
                assert answers[case] == result

    # A call with the same arguments occurs only once
    assert len(lst) == len(cases)


def test_cache_clear():
    calls = []

    @moka_py.cached()
    def f(x: int, y: int) -> float:
        calls.append((x, y))
        return x / y

    f(1, 2)
    f(1, 2)
    f.cache_clear()
    f(1, 2)
    assert len(calls) == 2


@pytest.mark.parametrize(("wait", "expected_calls"), [
    (True, 1),
    (False, 5)
])
def test_wait_concurrent(wait, expected_calls):
    calls = []

    @moka_py.cached(wait_concurrent=wait)
    def f(x: int, y: int) -> float:
        calls.append((x, y))
        sleep(0.1)
        return x / y

    def target():
        assert f(12, 3) == 4.0

    threads = [Thread(target=target) for _ in range(5)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()

    assert len(calls) == expected_calls



from typing import TypeVar, Optional, Generic, Hashable, Union, Callable, Any


K = TypeVar("K", bound=Hashable)
V = TypeVar("V")
Fn = TypeVar("Fn", bound=Callable[..., Any])


class Moka(Generic[K, V]):
    def __init__(
            self,
            capacity: int,
            ttl: Optional[Union[int, float]] = None,
            tti: Optional[Union[int, float]] = None,
    ): ...

    def set(self, key: K, value: V) -> None: ...

    def get(self, key: K) -> Optional[V]: ...

    def get_with(self, key: K, initializer: Callable[[], V]) -> V:
        """
        Lookups for a key in the cache and only if there is no value set, calls the `initializer`
        function to set the key's value.
        If multiple threads call `get_with` with the same key, only one of them calls
        `initializer`, and the others wait until the value is set.
        """

    def remove(self, key: K) -> Optional[V]: ...

    def clear(self) -> None: ...

    def count(self) -> int: ...


def cached(
        maxsize: int = 128,
        typed: bool = False,
        *,
        ttl: Optional[Union[int, float]] = None,
        tti: Optional[Union[int, float]] = None,
        wait_concurrent: bool = False,
) -> Callable[[Fn], Fn]:
    ...



import asyncio
from time import monotonic, sleep
import threading
import moka_py


async def test_tti():
    moka = moka_py.Moka(128, tti=0.2)
    value = {"foo": "bar"}
    moka.set("hello", value)
    assert moka.get("hello") is value
    assert moka.get("hello") is value
    await asyncio.sleep(0.2)
    assert moka.get("hello") is None


async def test_ttl_and_tti():
    ttl = 0.8
    tti = 0.3
    moka = moka_py.Moka(128, ttl=ttl, tti=tti)
    value = {"foo": "bar"}
    moka.set("hello", value)
    start = monotonic()
    while monotonic() - start <= ttl:
        assert moka.get("hello") is value
        # * 0.5 is a making sure that TTI is bumped in time
        # since asyncio.sleep only guarantees it won't be wakened up before
        # the specified time
        await asyncio.sleep(tti * 0.5)
    assert moka.get("hello") is None


def test_eviction():
    size = 128
    moka = moka_py.Moka(size)
    keys = list(range(500))
    assert len(keys) > size

    for key in keys:
        moka.set(key, key)

    got = []
    for key in keys:
        v = moka.get(key)
        if v is not None:
            got.append(v)

    assert len(got) == size


def test_remove():
    moka = moka_py.Moka(128)
    moka.set("hello", "world")
    assert moka.remove("hello") == "world"
    assert moka.get("hello") is None


def test_get_with():
    moka = moka_py.Moka(128)
    calls = []

    def init():
        calls.append(1)
        sleep(0.2)
        return "world"

    def target():
        res = moka.get_with("hello", init)
        assert res == "world"

    t1 = threading.Thread(target=target)
    t2 = threading.Thread(target=target)

    t1.start()
    t2.start()

    t1.join()
    t2.join()

    assert len(calls) == 1
